@@ -1,16 +1,23 @@
 (in-package :closer-mop)
 
-;; This is a useful utility function.
+;; Some utility functions.
 
 (defun required-args (lambda-list &optional (collector #'identity))
   (loop for arg in lambda-list
         until (member arg lambda-list-keywords)
         collect (funcall collector arg)))
 
+(defun ensure-finalized (class &optional (errorp t))
+  (if (typep class 'class)
+    (unless (class-finalized-p class)
+      (finalize-inheritance class))
+    (when errorp (error "~S is not a class." class)))
+  class)
+
 ;; We need a new standard-generic-function for various things.
 
 (cl:defclass standard-generic-function (cl:standard-generic-function)
-  ()
+  ((initial-methods :initform '()))
   (:metaclass clos:funcallable-standard-class))
 
 ;; The following ensures that the new standard-generic-function is used.
@@ -27,9 +34,9 @@
                 "~S is already fbound, but not as a generic function." name)
         (fmakunbound name))))
   (if (fboundp name)
-      (let ((function (fdefinition name)))
-        (apply #'ensure-generic-function-using-class
-               function name args))
+    (let ((function (fdefinition name)))
+      (apply #'ensure-generic-function-using-class
+             function name args))
     (apply #'ensure-generic-function-using-class nil name
            :generic-function-class generic-function-class
            args)))
@@ -53,9 +60,8 @@
            ((class standard-class)
             (superclass cl:standard-class))
   (or (when (eq (class-of class) (find-class 'standard-class))
-        (member (class-of superclass)
-                (list (find-class 'cl:standard-class)
-                      (find-class 'standard-class))))
+        (or (eq (class-of superclass) (find-class 'cl:standard-class))
+            (eq (class-of superclass) (find-class 'standard-class))))
       (call-next-method)
       (when (eq (class-of superclass) (find-class 'cl:standard-class))
         (validate-superclass class (class-prototype (find-class 'standard-class))))))
@@ -69,7 +75,7 @@
     
 (defmacro defclass (name (&rest supers) &body options)
   (if (member :metaclass options :key #'car)
-      `(cl:defclass ,name ,supers ,@options)
+    `(cl:defclass ,name ,supers ,@options)
     `(cl:defclass ,name ,supers ,@options
        (:metaclass standard-class))))
 
@@ -84,16 +90,22 @@
            ((class funcallable-standard-class)
             (superclass clos:funcallable-standard-class))
   (or (when (eq (class-of class) (find-class 'funcallable-standard-class))
-        (member (class-of superclass)
-                (list (find-class 'clos:funcallable-standard-class)
-                      (find-class 'funcallable-standard-class))))
+        (or (eq (class-of superclass) (find-class 'clos:funcallable-standard-class))
+            (eq (class-of superclass) (find-class 'funcallable-standard-class))))
       (call-next-method)
       (when (eq (class-of superclass) (find-class 'clos:funcallable-standard-class))
         (validate-superclass class (class-prototype (find-class 'funcallable-standard-class))))))
 
+#+lispworks5.0
+(cl:defmethod validate-superclass
+           ((class funcallable-standard-class)
+            (superclass (eql (find-class 'funcallable-standard-object))))
+  t)
+
 ;; We also need a new funcallable-standard-object because the default one
 ;; is not an instance of clos:funcallable-standard-class.
 
+#-lispworks5.0
 (cl:defclass funcallable-standard-object (clos:funcallable-standard-object)
   ()
   (:metaclass clos:funcallable-standard-class))
@@ -101,19 +113,20 @@
 ;; The following code ensures that possibly incorrect lists of direct
 ;; superclasses are corrected.
 
+#-lispworks5.0
 (defun modify-superclasses (direct-superclasses &optional (standardp t))
   (if (null direct-superclasses)
-      (list (if standardp
-                (find-class 'standard-object)
-              (find-class 'funcallable-standard-object)))
+    (list (if standardp
+            (find-class 'standard-object)
+            (find-class 'funcallable-standard-object)))
     (let ((standard-object (if standardp
-                               (find-class 'standard-object)
+                             (find-class 'standard-object)
                              (find-class 'clos:funcallable-standard-object))))
       (if (eq (car (last direct-superclasses)) standard-object)
-          (if standardp
-              direct-superclasses
-            (append (butlast direct-superclasses)
-                    (list (find-class 'funcallable-standard-object))))
+        (if standardp
+          direct-superclasses
+          (append (butlast direct-superclasses)
+                  (list (find-class 'funcallable-standard-object))))
         (remove standard-object direct-superclasses)))))
 
 ;; During class re/initialization, we take care of the following things:
@@ -123,63 +136,82 @@
 
 (cl:defmethod initialize-instance :around
   ((class standard-class) &rest initargs
-   &key (direct-superclasses ()))
+   #-lispworks5.0 &key
+   #-lispworks5.0 (direct-superclasses ()))
   (declare (dynamic-extent initargs))
   (apply #'call-next-method class
-         :direct-superclasses (modify-superclasses direct-superclasses)
+         #-lispworks5.0 :direct-superclasses
+         #-lispworks5.0 (modify-superclasses direct-superclasses)
          :optimize-slot-access nil
          initargs))
 
 (cl:defmethod reinitialize-instance :around
   ((class standard-class) &rest initargs
-   &key  (direct-superclasses () direct-superclasses-p))
+   #-lispworks5.0 &key
+   #-lispworks5.0 (direct-superclasses () direct-superclasses-p))
   (declare (dynamic-extent initargs))
-  (when direct-superclasses-p
-    (setq direct-superclasses (modify-superclasses direct-superclasses))
-    (loop for superclass in (copy-list (class-direct-superclasses class))
-          unless (member superclass direct-superclasses)
-          do (remove-direct-subclass superclass class)))
-  (if direct-superclasses-p
+  #-lispworks5.0
+  (progn
+    (when direct-superclasses-p
+      (setq direct-superclasses (modify-superclasses direct-superclasses))
+      (loop for superclass in (copy-list (class-direct-superclasses class))
+            unless (member superclass direct-superclasses)
+            do (remove-direct-subclass superclass class)))
+    (if direct-superclasses-p
       (apply #'call-next-method class
              :direct-superclasses direct-superclasses
              :optimize-slot-access nil
              initargs)
-    (apply #'call-next-method class
-           :optimize-slot-access nil
-           initargs)))
+      (apply #'call-next-method class
+             :optimize-slot-access nil
+             initargs)))
+  #+lispworks5.0
+  (apply #'call-next-method class
+         :optimize-slot-access nil
+         initargs))
 
 (cl:defmethod initialize-instance :around
   ((class funcallable-standard-class) &rest initargs
-   &key (direct-superclasses ()))
+   #-lispworks5.0 &key
+   #-lispworks5.0 (direct-superclasses ()))
   (declare (dynamic-extent initargs))
   (apply #'call-next-method class
-         :direct-superclasses (modify-superclasses direct-superclasses nil)
+         #-lispworks5.0 :direct-superclasses
+         #-lispworks5.0 (modify-superclasses direct-superclasses nil)
          :optimize-slot-access nil
          initargs))
 
 (cl:defmethod reinitialize-instance :around
   ((class funcallable-standard-class) &rest initargs
-   &key (direct-superclasses () direct-superclasses-p))
+   #-lispworks5.0 &key
+   #-lispworks5.0 (direct-superclasses () direct-superclasses-p))
   (declare (dynamic-extent initargs))
-  (when direct-superclasses-p
-    (setq direct-superclasses (modify-superclasses direct-superclasses nil))
-    (loop for superclass in (copy-list (class-direct-superclasses class))
-          unless (member superclass direct-superclasses)
-          do (remove-direct-subclass superclass class)))
-  (if direct-superclasses-p
+  #-lispworks5.0
+  (progn
+    (when direct-superclasses-p
+      (setq direct-superclasses (modify-superclasses direct-superclasses nil))
+      (loop for superclass in (copy-list (class-direct-superclasses class))
+            unless (member superclass direct-superclasses)
+            do (remove-direct-subclass superclass class)))
+    (if direct-superclasses-p
       (apply #'call-next-method class
              :direct-superclasses direct-superclasses
              :optimize-slot-access nil
              initargs)
-    (apply #'call-next-method class
-           :optimize-slot-access nil
-           initargs)))
+      (apply #'call-next-method class
+             :optimize-slot-access nil
+             initargs)))
+  #+lispworks5.0
+  (apply #'call-next-method class
+         :optimize-slot-access nil
+         initargs))
 
 ;; The following is necessary for forward-referenced-classes.
 ;; Since we replace the original funcallable-standard-object with
 ;; a new one, we have to prevent LispWorks from trying to use
 ;; the original one when forward-ferenced-classes are resolved.
 
+#-lispworks5.0
 (cl:defmethod change-class :around
   ((class forward-referenced-class)
    (new-class funcallable-standard-class)
@@ -203,7 +235,7 @@
                      :test #'eq
                      :key #'slot-definition-name)))
     (if slotd
-        (slot-value-using-class class object slotd)
+      (slot-value-using-class class object slotd)
       (slot-missing class object slot 'slot-value))))
 
 (cl:defmethod slot-value-using-class
@@ -223,8 +255,8 @@
                      :test #'eq
                      :key #'slot-definition-name)))
     (if slotd
-        (setf (slot-value-using-class class object slotd)
-              new-value)
+      (setf (slot-value-using-class class object slotd)
+            new-value)
       (slot-missing class object slot 'setf new-value))))
 
 (cl:defmethod (setf slot-value-using-class)
@@ -245,7 +277,7 @@
                      :test #'eq
                      :key #'slot-definition-name)))
     (if slotd
-        (slot-boundp-using-class class object slotd)
+      (slot-boundp-using-class class object slotd)
       (slot-missing class object slot 'slot-boundp))))
 
 (cl:defmethod slot-boundp-using-class
@@ -265,7 +297,7 @@
                      :test #'eq
                      :key #'slot-definition-name)))
     (if slotd
-        (slot-makunbound-using-class class object slotd)
+      (slot-makunbound-using-class class object slotd)
       (slot-missing class object slot 'slot-makunbound))))
 
 (cl:defmethod slot-makunbound-using-class
@@ -287,16 +319,16 @@
 (cl:defgeneric eql-specializer-object (eql-specializer)
   (:method ((cons cons))
    (if (clos:eql-specializer-p cons)
-       (cadr cons)
+     (cadr cons)
      (error "~S is not an eql-specializer." cons))))
 
 (defun intern-eql-specializer (object)
   `(eql ,object))
 
 (cl:defclass eql-specializer* (metaobject)
-  ((object :reader eql-specializer-object
-           :initarg eso
-           :initform (error "Use intern-eql-specializer to create eql-specializers."))
+  ((obj :reader eql-specializer-object
+        :initarg eso
+        :initform (error "Use intern-eql-specializer to create eql-specializers."))
    (direct-methods :reader specializer-direct-methods
                    :accessor es-direct-methods
                    :initform ())))
@@ -330,6 +362,7 @@
 
 ;; The following method ensures that remove-method is called.
 
+#-lispworks5.0
 (cl:defmethod add-method :before ((gf standard-generic-function) (method method))
   (when-let (old-method (find-method gf (method-qualifiers method)
                                      (method-specializers method) nil))
@@ -340,24 +373,28 @@
 
 (cl:defmethod add-method :after ((gf standard-generic-function) (method method))
   (loop for specializer in (method-specializers method)
+        if (consp specializer)
         do (add-direct-method
-            (if (consp specializer)
-                (intern-eql-specializer*
-                 (eql-specializer-object specializer))
-              specializer)
-            method))
+            (intern-eql-specializer*
+             (eql-specializer-object specializer))
+            method)
+        #-lispworks5.0 else
+        #-lispworks5.0 do
+        #-lispworks5.0 (add-direct-method specializer method))
   #+lispworks4.3
   (map-dependents
    gf (lambda (dep) (update-dependent gf dep 'add-method method))))
 
 (cl:defmethod remove-method :after ((gf standard-generic-function) (method method))
   (loop for specializer in (method-specializers method)
+        if (consp specializer)
         do (remove-direct-method
-            (if (consp specializer)
-                (intern-eql-specializer*
-                 (eql-specializer-object specializer))
-              specializer)
-            method))
+            (intern-eql-specializer*
+             (eql-specializer-object specializer))
+            method)
+        #-lispworks5.0 else
+        #-lispworks5.0 do
+        #-lispworks5.0 (remove-direct-method specializer method))
   #+lispworks4.3
   (map-dependents
    gf (lambda (dep) (update-dependent gf dep 'remove-method method))))
@@ -382,17 +419,15 @@
        lambda-expression
      (declare (ignore lambda))
      (loop with documentation = :unbound
-           for tail on body
-           while (or (stringp (car tail))
-                     (and (consp (car tail))
-                          (eq (caar tail) 'declare)))
-           if (stringp (car tail))
+           for (car . cdr) = body then cdr
+           while (or (stringp car)
+                     (and (consp car) (eq (car car) 'declare)))
+           if (stringp car)
            do (setf documentation
-                    (if (eq documentation :unbound)
-                        (car tail)
+                    (if (eq documentation :unbound) car
                       (error "Too many documentation strings in lambda expression ~S."
                              lambda-expression)))
-           else append (loop for declaration in (cdar tail) 
+           else append (loop for declaration in (cdr car) 
                              if (eq (car declaration) 'ignore)
                              collect `(ignorable ,@(cdr declaration))
                              and collect `(dynamic-extent ,@(cdr declaration))
@@ -401,9 +436,9 @@
                        (method-lambda method-args)
                        (clos:make-method-lambda
                         gf method args declarations
-                        `(progn ,@tail) env)
+                        `(progn ,car ,@cdr) env)
                      (if (eq documentation :unbound)
-                         (return (values method-lambda method-args))
+                       (return (values method-lambda method-args))
                        (return (values
                                 `(lambda ,(cadr method-lambda)
                                    ,documentation
@@ -438,41 +473,38 @@
         until (member arg stop-keywords)
         collect arg into gf-lambda-list
         finally (return (if (member arg '(&rest &key))
-                            (append gf-lambda-list '(&key))
+                          (append gf-lambda-list '(&key))
                           gf-lambda-list))))
 
 ;; The defmethod macro is needed in order to ensure that make-method-lambda
 ;; is called. (Unfortunately, this doesn't work in the other CL implementations.)
 
 (defmacro defmethod (&whole form name &body body &environment env)
-  (loop for tail on body
+  (loop for tail = body then (cdr tail)
         until (listp (car tail))
         collect (car tail) into qualifiers
         finally
         (destructuring-bind
             ((&rest specialized-args) &body body) tail
           (loop with documentation = :unbound
-                for tail on body
-                while (or (stringp (car tail))
-                          (and (consp (car tail))
-                               (eq (caar tail) 'declare)))
-                if (stringp (car tail))
+                for (car . cdr) = body then cdr
+                while (or (stringp car)
+                          (and (consp car) (eq (car car) 'declare)))
+                if (stringp car)
                 do (setq documentation
-                         (if (eq documentation :unbound)
-                             (car tail)
+                         (if (eq documentation :unbound) car
                            (error "Too many documentation strings for defmethod form ~S." form)))
-                else append (cdar tail) into declarations
+                else append (cdr car) into declarations
                 finally
                 (let* ((lambda-list (extract-lambda-list specialized-args))
                        (gf-lambda-list (create-gf-lambda-list lambda-list))
                        (gf (if (fboundp name)
-                               (ensure-generic-function name)
+                             (ensure-generic-function name)
                              (ensure-generic-function name :lambda-list gf-lambda-list)))
                        (method-class (generic-function-method-class gf))
                        (lambda-expression `(lambda ,lambda-list
                                              (declare ,@declarations)
-                                             (block ,name
-                                               ,@tail))))
+                                             (block ,name ,car ,@cdr))))
                   (if (equal (compute-applicable-methods
                               #'make-method-lambda
                               (list gf (class-prototype method-class)
@@ -484,7 +516,7 @@
                                           (find-class 't)
                                           (find-class 't))
                                     nil)))
-                      (return-from defmethod `(cl:defmethod ,@(rest form)))
+                    (return-from defmethod `(cl:defmethod ,@(rest form)))
                     (multiple-value-bind
                         (method-lambda method-args)
                         (make-method-lambda
@@ -493,7 +525,7 @@
                       (with-unique-names (gf method)
                         (return-from defmethod
                           `(let ((,gf (if (fboundp ',name)
-                                          (ensure-generic-function ',name)
+                                        (ensure-generic-function ',name)
                                         (ensure-generic-function
                                          ',name :lambda-list ',gf-lambda-list)))
                                  (,method
@@ -529,22 +561,23 @@
 ;; is used by default. It also ensures that make-method-lambda is called
 ;; for the default methods, by expanding into defmethod forms.
 
-(defvar *gf-initial-methods* (make-hash-table :test #'equal))
-
 (defmacro defgeneric (&whole form name (&rest args) &body options)
   (unless (every #'consp options)
     (error "Illegal generic functions options in defgeneric form ~S." form))
   `(progn
-     (loop for method in (gethash ',name *gf-initial-methods*)
-           do (remove-method (method-generic-function method) method))
+     (let ((generic-function (ignore-errors (fdefinition ',name))))
+       (when (and generic-function (typep generic-function 'standard-generic-function))
+         (loop for method in (slot-value generic-function 'initial-methods)
+               do (remove-method generic-function method))))
      (cl:defgeneric ,name ,args
        ,@(remove :method options :key #'car :test #'eq)
        ,@(unless (member :generic-function-class options :key #'car :test #'eq)
            '((:generic-function-class standard-generic-function))))
-     (setf (gethash ',name *gf-initial-methods*)
-           (list ,@(loop for method-spec in (remove :method options :key #'car :test-not #'eq)
-                         collect `(defmethod ,name ,@(cdr method-spec)))))
-     (fdefinition ',name)))
+     (let ((generic-function (fdefinition ',name)))
+       (setf (slot-value generic-function 'initial-methods)
+             (list ,@(loop for method-spec in (remove :method options :key #'car :test-not #'eq)
+                           collect `(defmethod ,name ,@(cdr method-spec)))))
+       generic-function)))
 
 ;; The following can be used in direct-slot-definition-class to get the correct initargs
 ;; for a slot. Use it like this:
